%
% Copyright (c) 2015, Yarpiz (www.yarpiz.com)
% All rights reserved. Please read the "license.txt" for license terms.
%
% Project Code: YPEA104
% Project Title: Ant Colony Optimization for Continuous Domains (ACOR)
% Publisher: Yarpiz (www.yarpiz.com)
% 
% Developer: S. Mostapha Kalami Heris (Member of Yarpiz Team)
% 
% Contact Info: sm.kalami@gmail.com, info@yarpiz.com
%
function [bestX, bestFx]=aco(fx,Lb,Ub,MaxPop,MaxIters)
%% Problem Definition

  nVars=length(Lb);             % Number of Decision Variables
  %% ACOR Parameters
  MaxStall = fix(MaxIters/20);
  lastIter = 0;
  nSample=4*MaxPop;         % Sample Size
  q=0.5;              % Intensification Factor (Selection Pressure)
  zeta=1;             % Deviation-Distance Ratio
  %% Initialization
  % Create Empty Individual Structure
  empty_individual.Position=[];
  empty_individual.Cost=[];
  % Create Population Matrix
  pop=repmat(empty_individual,MaxPop,1);
  bestFx = 1e99;
  % Initialize Population Members
  for i=1:MaxPop
  
    % Create Random Solution
    for j=1:nVars
      pop(i).Position(j) = Lb(j) + (Ub(j)-Lb(j))*rand;
    end
  
    % Evaluation
    pop(i).Cost=fx(pop(i).Position);
  end
  % Sort Population
  [~, SortOrder]=sort([pop.Cost]);
  pop=pop(SortOrder);
  % Update Best Solution Ever Found
  bestFx = pop(1).Cost;
  bestX = pop(1).Position;
  fprintf("Iter 0, Fx = %e, X=[", bestFx);
  fprintf("%f, ", bestX(1:nVars-1));
  fprintf("%f]\n", bestX(nVars));
  
  % Solution Weights
  w=1/(sqrt(2*pi)*q*MaxPop)*exp(-0.5*(((1:MaxPop)-1)/(q*MaxPop)).^2);
  % Selection Probabilities
  p=w/sum(w);
  
  %% ACOR Main Loop
  for iter=1:MaxIters
    if (iter - lastIter) > MaxStall
      lastIter = iter;
  	  for i=1:MaxPop
        % Create Random Solution
    	  for j=1:nVars
          pop(i).Position(j) = Lb(j) + (Ub(j)-Lb(j))*rand;
        end
  
        % Evaluation
        pop(i).Cost=fx(pop(i).Position);
      end
      % Sort Population
      [~, SortOrder]=sort([pop.Cost]);
      pop=pop(SortOrder);
      % Update Best Solution Ever Found
      if bestFx > pop(1).Cost
        bestFx = pop(1).Cost;
        bestX = pop(1).Position;
      end
    end
  
    % Means
    s=zeros(MaxPop,nVars);
    for l=1:MaxPop
      s(l,:)=pop(l).Position;
    end
  
    % Standard Deviations
    sigma=zeros(MaxPop,nVars);
    for l=1:MaxPop
      D=0;
      for r=1:MaxPop
        D=D+abs(s(l,:)-s(r,:));
      end
      sigma(l,:)=zeta*D/(MaxPop-1);
    end
  
    % Create New Population Array
    newpop=repmat(empty_individual,nSample,1);
    for t=1:nSample
  
      % Initialize Position Matrix
      for j=1:nVars
        newpop(t).Position(j) = Lb(j) + (Ub(j)-Lb(j))*rand;
      end
  
      % Solution Construction
      for i=1:nVars
  
        % Select Gaussian Kernel
        l=RouletteWheelSelection(p);
  
        % Generate Gaussian Random Variable
        newpop(t).Position(i)=s(l,i)+sigma(l,i)*randn;
        if newpop(t).Position(i) < Lb(i) || ...
            newpop(t).Position(i) > Ub(i)
          newpop(t).Position(i) = Lb(i) + (Ub(i)-Lb(i))*rand;
        end
      end
  
      % Evaluation
      newpop(t).Cost=fx(newpop(t).Position);
      
    end
  
    % Merge Main Population (Archive) and New Population (Samples)
    pop=[pop
      newpop]; %#ok
  
    % Sort Population
    [~, SortOrder]=sort([pop.Cost]);
    pop=pop(SortOrder);
  
    % Delete Extra Members
    pop=pop(1:MaxPop);
    if bestFx > pop(1).Cost
      lastIter = iter;
      bestFx = pop(1).Cost;
      bestX = pop(1).Position;
      % Show Iteration Information
      fprintf("Iter %d, Fx = %e, X=[", iter, bestFx);
      fprintf("%f, ", bestX(1:nVars-1));
      fprintf("%f]\n", bestX(nVars));
      % disp(['Iteration ' num2str(iter) ': Best Cost = ' num2str(BestCost(iter))]);
    end
end
