% -----------------------------------------------------------------
% Cuckoo Search (CS) algorithm by Xin-She Yang and Suash Deb      %
% Programmed by Xin-She Yang at Cambridge University              %
% Programming dates: Nov 2008 to June 2009                        %
% Last revised: Dec  2009   (simplified version for demo only)    %
% -----------------------------------------------------------------
% Papers -- Citation Details:
% 1) X.-S. Yang, S. Deb, Cuckoo search via Levy flights,
% in: Proc. of World Congress on Nature & Biologically Inspired
% Computing (NaBIC 2009), December 2009, India,
% IEEE Publications, USA,  pp. 210-214 (2009).
% http://arxiv.org/PS_cache/arxiv/pdf/1003/1003.1594v1.pdf 
% 2) X.-S. Yang, S. Deb, Engineering optimization by cuckoo search,
% Int. J. Mathematical Modelling and Numerical Optimisation, 
% Vol. 1, No. 4, 330-343 (2010). 
% http://arxiv.org/PS_cache/arxiv/pdf/1005/1005.2908v2.pdf
% ----------------------------------------------------------------%
% This demo program only implements a standard version of         %
% Cuckoo Search (CS), as the Levy flights and generation of       %
% new solutions may use slightly different methods.               %
% The pseudo code was given sequentially (select a cuckoo etc),   %
% but the implementation here uses Matlab's vector capability,    %
% which results in neater/better codes and shorter running time.  % 
% This implementation is different and more efficient than the    %
% the demo code provided in the book by 
%    "Yang X. S., Nature-Inspired Metaheuristic Algoirthms,       % 
%     2nd Edition, Luniver Press, (2010).                 "       %
% --------------------------------------------------------------- %

% =============================================================== %
% Notes:                                                          %
% Different implementations may lead to slightly different        %
% behavour and/or results, but there is nothing wrong with it,    %
% as this is the nature of random walks and all metaheuristics.   %
% -----------------------------------------------------------------

% Note: Very good algorithm

function [bestX,bestFx]=cuckoo_search(fx,Lb,Ub,MaxPop,MaxIters,Toler)

  if nargin < 6, Toler = 1e-15; end
  Toler = abs(Toler);
  % Discovery rate of alien eggs/solutions
  pa=0.25;

  %% Simple bounds of the search domain
  nVars=length(Lb); 
  pop = zeros(MaxPop,nVars);
  pop = zeros(MaxPop,nVars);
  % Random initial solutions
  for i=1:MaxPop
    pop(i,:)=Lb+(Ub-Lb).*rand(size(Lb));
  end

  % Get the current best
  fitness=10^10*ones(MaxPop,1);
  [bestFx,bestX,pop,fitness]=get_best_nest(fx,pop,pop,fitness);

  N_iter=0;
  
  %% Starting iterations
  for iter=1:MaxIters
      
      % Generate new solutions (but keep the current best)
       new_nest=get_cuckoos(pop,bestX,Lb,Ub);   
       [fnew,best,pop,fitness]=get_best_nest(fx,pop,new_nest,fitness);
      % Update the counter
        N_iter=N_iter+MaxPop; 
      % Discovery and randomization
        new_nest=empty_nests(pop,Lb,Ub,pa) ;

      % Evaluate this set of solutions
        [fnew,best,pop,fitness]=get_best_nest(fx,pop,new_nest,fitness);
      % Update the counter again
        N_iter=N_iter+MaxPop;
      % Find the best objective so far  
      if fnew < bestFx
          bestFx = fnew;
          bestX = best;

          fprintf("iter: %d, Fx = %e, X=[", iter, bestFx);
          fprintf("%f, ", bestX(1:nVars-1));
          fprintf("%f]\n", bestX(nVars));
      end
      
      if abs(bestFx) <= Toler
        break;
      end
  end %% End of iterations
end

%% --------------- All subfunctions are list below ------------------
%% Get cuckoos by ramdom walk
function pop=get_cuckoos(pop,best,Lb,Ub)
  % Levy flights
  MaxPop=size(pop,1);
  % Levy exponent and coefficient
  % For details, see equation (2.21), Page 16 (chapter 2) of the book
  % X. S. Yang, Nature-Inspired Metaheuristic Algorithms, 2nd Edition, Luniver Press, (2010).
  beta=3/2;
  sigma=(gamma(1+beta)*sin(pi*beta/2)/(gamma((1+beta)/2)*beta*2^((beta-1)/2)))^(1/beta);

  for j=1:MaxPop
      s=pop(j,:);
      % This is a simple way of implementing Levy flights
      % For standard random walks, use step=1;
      %% Levy flights by Mantegna's algorithm
      u=randn(size(s))*sigma;
      v=randn(size(s));
      step=u./abs(v).^(1/beta);

      % In the next equation, the difference factor (s-best) means that 
      % when the solution is the best solution, it remains unchanged.     
      stepsize=0.01*step.*(s-best);
      % Here the factor 0.01 comes from the fact that L/100 should the typical
      % step size of walks/flights where L is the typical lenghtscale; 
      % otherwise, Levy flights may become too aggresive/efficient, 
      % which makes new solutions (even) jump out side of the design domain 
      % (and thus wasting evaluations).
      % Now the actual random walks or flights
      s=s+stepsize.*randn(size(s));
     % Apply simple bounds/limits
     pop(j,:)=simplebounds(s,Lb,Ub);
  end
end

%% Find the current best pop
function [bestFx,best,pop,fitness]=get_best_nest(fx, pop,newnest,fitness)
  % Evaluating all new solutions
  for j=1:size(pop,1)
      fnew=fx(newnest(j,:));
      if fnew<=fitness(j)
         fitness(j)=fnew;
         pop(j,:)=newnest(j,:);
      end
  end
  % Find the current best
  [bestFx,K]=min(fitness) ;
  best=pop(K,:);
end

%% Replace some nests by constructing new solutions/nests
function new_nest=empty_nests(pop,Lb,Ub,pa)
  % A fraction of worse nests are discovered with a probability pa
  MaxPop=size(pop,1);
  nVars=size(pop,2);
  % Discovered or not -- a status vector
  K=rand(size(pop))>pa;

  % In the real world, if a cuckoo's egg is very similar to a host's eggs, then 
  % this cuckoo's egg is less likely to be discovered, thus the fitness should 
  % be related to the difference in solutions.  Therefore, it is a good idea 
  % to do a random walk in a biased way with some random step sizes.  
  %% New solution by biased/selective random walks
  stepsize = zeros(MaxPop,nVars);
  while norm(stepsize) == 0
    stepsize=rand*(pop(randperm(MaxPop),:)-pop(randperm(MaxPop),:));
  end
  new_nest=pop+stepsize.*K;
  for j=1:size(new_nest,1)
    s=new_nest(j,:);
    new_nest(j,:)=simplebounds(s,Lb,Ub);  
  end
end

% Application of simple constraints
function s=simplebounds(s,Lb,Ub)
  for i=1:length(s)
    if s(i) < Lb(i) || s(i) > Ub(i)
      s(i) = Lb(i) + (Ub(i) - Lb(i))*rand;
    end
  end
end


