% -----------------------------------------------------------------
% Cuckoo Search (CS) algorithm by Xin-She Yang and Suash Deb      %
% Programmed by Xin-She Yang at Cambridge University              %
% Programming dates: Nov 2008 to June 2009                        %
% Last revised: Dec  2009   (simplified version for demo only)    %
% -----------------------------------------------------------------
% Papers -- Citation Details:
% 1) X.-S. Yang, S. Deb, Cuckoo search via Levy flights,
% in: Proc. of World Congress on Nature & Biologically Inspired
% Computing (NaBIC 2009), December 2009, India,
% IEEE Publications, USA,  pp. 210-214 (2009).
% http://arxiv.org/PS_cache/arxiv/pdf/1003/1003.1594v1.pdf 
% 2) X.-S. Yang, S. Deb, Engineering optimization by cuckoo search,
% Int. J. Mathematical Modelling and Numerical Optimisation, 
% Vol. 1, No. 4, 330-343 (2010). 
% http://arxiv.org/PS_cache/arxiv/pdf/1005/1005.2908v2.pdf
% ----------------------------------------------------------------%
% This demo program only implements a standard version of         %
% Cuckoo Search (CS), as the Levy flights and generation of       %
% new solutions may use slightly different methods.               %
% The pseudo code was given sequentially (select a cuckoo etc),   %
% but the implementation here uses Matlab's vector capability,    %
% which results in neater/better codes and shorter running time.  % 
% This implementation is different and more efficient than the    %
% the demo code provided in the book by 
%    "Yang X. S., Nature-Inspired Metaheuristic Algoirthms,       % 
%     2nd Edition, Luniver Press, (2010).                 "       %
% --------------------------------------------------------------- %

% =============================================================== %
% Notes:                                                          %
% Different implementations may lead to slightly different        %
% behavour and/or results, but there is nothing wrong with it,    %
% as this is the nature of random walks and all metaheuristics.   %
% -----------------------------------------------------------------

% -----------------------------------------------------------------
% Note: Code altered by Namir Shammas. This version uses an augmented 
% version of the matrix nest, where the last column has the function value
% for the data in that row. Also, the code updates the second half of the
% population in matrix nest as long as parameter pa is less than a random 
% number. The values of optimization parameters vary between 0.1 and 0.5, 
% based on the main iterations loop counter.
%
% -----------------------------------------------------------------

% Note: Very good algorithm

function [bestX,bestFx]=cuckoo_searchEx4(fx,Lb,Ub,MaxPop,MaxIters,Toler)
  if nargin < 6, Toler = 1e-15; end
  % Number of nests (or different solutions)
  n=MaxPop;

  %% Simple bounds of the search domain
  nVars = length(Lb); 
  m = nVars + 1;
  nest = zeros(n,m); % additional column for function values
  % Random initial solutions
  for i=1:n
    nest(i,1:nVars)=Lb+(Ub-Lb).*rand(1,nVars);
    nest(i,m) = fx(nest(i,1:nVars));
  end

  % sort nests, such that best nest is at index 1
  nest = sortrows(nest,m);
  
  %% Starting iterations
  for iter=1:MaxIters
      pa = 0.5 - 0.4*(iter - 1)/(MaxIters - 1);

      % Generate new solutions (but keep the current best)
       new_nest=get_cuckoos(nest,Lb,Ub);   
       [nest]=get_best_nest(fx,nest,new_nest,iter);

      % Discovery and randomization
        new_nest=empty_nests(fx,nest,Lb,Ub,pa) ;

      % Evaluate this set of solutions
        [nest]=get_best_nest(fx,nest,new_nest,iter);

      if abs(nest(1,m)) <= Toler
        break;
      end
  end %% End of iterations
  bestX = nest(1, 1:nVars);
  bestFx = nest(1, m);
end

%% --------------- All subfunctions are list below ------------------
%% Get cuckoos by ramdom walk
function nest=get_cuckoos(nest,Lb,Ub)
  % Levy flights
  [n, m] = size(nest);
  nVars = m - 1;
  % Levy exponent and coefficient
  % For details, see equation (2.21), Page 16 (chapter 2) of the book
  % X. S. Yang, Nature-Inspired Metaheuristic Algorithms, 2nd Edition, Luniver Press, (2010).
  beta=3/2;
  sigma=(gamma(1+beta)*sin(pi*beta/2)/(gamma((1+beta)/2)*beta*2^((beta-1)/2)))^(1/beta);

  for j=1:n
      s=nest(j,1:nVars);
      % This is a simple way of implementing Levy flights
      % For standard random walks, use step=1;
      %% Levy flights by Mantegna's algorithm
      u=randn(size(s))*sigma;
      v=randn(size(s));
      step=u./abs(v).^(1/beta);

      % In the next equation, the difference factor (s-best) means that 
      % when the solution is the best solution, it remains unchanged.     
      stepsize=0.01*step.*(s-nest(1,1:nVars));
      % Here the factor 0.01 comes from the fact that L/100 should the typical
      % step size of walks/flights where L is the typical lenghtscale; 
      % otherwise, Levy flights may become too aggresive/efficient, 
      % which makes new solutions (even) jump out side of the design domain 
      % (and thus wasting evaluations).
      % Now the actual random walks or flights
      s=s+stepsize.*randn(size(s));
     % Apply simple bounds/limits
     nest(j,1:nVars) = simplebounds(s,Lb,Ub);
  end
end

%% Find the current best nest
function [nest]=get_best_nest(fx,nest,newnest,iter)
% Note: newnest is a matric that n by nVars, while nest is a matrix that us n by nVars+1.
  % Evaluating all new solutions
  [n, m] = size(nest);
  nVars = m - 1;
  bestFx = nest(1,m);
  for j=2:n
      fnew=fx(newnest(j,1:nVars));
      if fnew<=nest(j,m)
         nest(j,m) = fnew;
         nest(j,1:nVars) = newnest(j,1:nVars);
      end
  end
  
  nest = sortrows(nest, m);    
  if bestFx > nest(1,m)
    fprintf("Iter: %d, Fx = %e, X=[", iter, nest(j,m));
    fprintf("%f, ", nest(j,1:nVars-1));
    fprintf("%f]\n", nest(j,nVars));          
  end
end

%% Replace some nests by constructing new solutions/nests
function new_nest=empty_nests(fx,nest,Lb,Ub,pa)
  % The lower population is replaced
  [n, m] = size(nest);
  nVars = m - 1;
  nby2 = fix(n/2);
  for i=1:nby2
    i1 = 1 + fix(nby2*rand);
    i2 = i1;
    while i1==i2
      i2 = 1 + fix(nby2*rand);
    end
    stepsize = rand*(nest(i1+nby2,1:nVars)-nest(i2++nby2,1:nVars));
    if rand>pa
      new_nest(i,1:nVars)=nest(i,1:nVars)+stepsize;
      new_nest(i,1:nVars)=simplebounds(new_nest(i,1:nVars),Lb,Ub);
      new_nest(i, m) = fx(new_nest(i,1:nVars));
    else
      new_nest(i,:) = nest(i+nby2,:);
    end
    new_nest(i,:) = simplebounds(new_nest(i,:),Lb,Ub);
  end  
   % new_nest = [nest(1:nby2,:) new_nest(1:nby2,:)];
   new_nest = [nest(1:nby2,:); new_nest(1:nby2,:)];
end

% Application of simple constraints
function s=simplebounds(s,Lb,Ub)
  nVars = length(s)-1;
  for i=1:nVars
    if s(i) < Lb(i) || s(i) > Ub(i)
      s(i) = Lb(i) + (Ub(i) - Lb(i))*rand;
    end
  end
end


