function [gBestX,gBestCost] = pso(fx,Lb,Ub,MaxPop,MaxIters,itersToReset)
% PSO implements particle swarm optimization.
%
% INPUT
% ======
% fx - handle of optimized function.
% Lb - array of low bound values.
% Ub - array of upper bound values.
% MaxPop - maximum population of swarm.
% MaxIters - maxumum number of iterations
% itersToReset - number of stalled iterations that lead to data reset.
%
% OUTPUT
% ======
% gBestX - array of best solutions.
% gBestCost - best optimized function value.
%
% Example
% =======
%
% >> [gBestX,gBestCost] = pso(@fx2,[0 0 0 0],[5 5 5 5],50,500)
%

  nVar = length(Lb);
  template.position = zeros(1,nVar);
  template.velocity = zeros(1,nVar);
  template.cost =  0;
  template.best_position = zeros(1,nVar);
  template.best_cost = 1e+99;
  
  c1 = 1.5;
  c2 = 2;
  w = 1;
  wdamp = 0.995;
  gBestCost = 1e+99;
  gBestX = zeros(1,nVar);

  % Initizialize population
  for i=1:MaxPop
    pop(i) = template;
    for j=1:nVar
      pop(i).position(j) = Lb(j) + (Ub(j) - Lb(j))*rand;
      pop(i).velocity(j) = 0;
    end
    
    pop(i).cost = fx(pop(i).position);
    
    if pop(i).cost < pop(i).best_cost
      pop(i).best_cost = pop(i).cost; 
      pop(i).best_position = pop(i).position;

    end 
    
    if pop(i).best_cost < gBestCost
      gBestCost = pop(i).best_cost;
      gBestX = pop(i).best_position;
    end 
  end

  fprintf('Best Fx = %e, ', gBestCost);  
  fprintf('Best X =[');
  fprintf('%f, ', gBestX(1:nVar-1));
  fprintf(' %f]\n', gBestX(nVar));
  randSearchSuccessCount = 0;
  lastIter = 0;
  % pso loop
  for iter = 1:MaxIters
    if gBestCost == 0, return; end
    if (iter - lastIter) > itersToReset
      fprintf("----------------------- Reset population ----------------------- \n")
      fprintf("AT iteration %d\n", iter);
      lastIter = iter;
      for i=1:MaxPop
        pop(i) = template;
        for j=1:nVar
          pop(i).position(j) = Lb(j) + (Ub(j) - Lb(j))*rand;
          pop(i).velocity(j) = 0;
        end
        
        pop(i).cost = fx(pop(i).position);
        
        if pop(i).cost < pop(i).best_cost
          pop(i).best_cost = pop(i).cost; 
          pop(i).best_position = pop(i).position;
        end 
        
        if pop(i).best_cost < gBestCost
          gBestCost = pop(i).best_cost;
          gBestX = pop(i).best_position;
        end 
      end      
    end

    for i = 1:MaxPop
      for j = 1:nVar
        pop(i).velocity(j) = w * pop(i).velocity(j) + ...
          c1 * rand * (pop(i).best_position(j) - pop(i).position(j)) + ...
          c2 * rand * (gBestX(j) - pop(i).position(j));
        pop(i).position(j) = pop(i).position(j) + pop(i).velocity(j);
        if pop(i).position(j) < Lb(j) || pop(i).position(j) > Ub(j)
          pop(i).position(j) = Lb(j) + (Ub(j) - Lb(j)) * rand;
        end 
      end
      
      pop(i).cost = fx(pop(i).position);
      
      if pop(i).cost < pop(i).best_cost 
        pop(i).best_cost = pop(i).cost;
        pop(i).best_position = pop(i).position;
        
        if pop(i).best_cost < gBestCost 
          lastIter = iter;
          gBestCost = pop(i).best_cost;
          km = 25;
          Lb2 = Lb;
          Ub2 = Ub; 
          for j = 1:nVar
            gBestX(j) = pop(i).best_position(j);
            if gBestX(j) > 0
              Lb2(j) = 0.85*gBestX(j);
              Ub2(j) = 1.15*gBestX(j);
            elseif gBestX(j) < 0
               Lb2(j) = 1.15*gBestX(j);
               Ub2(j) = 0.85*gBestX(j);
            else
               Lb2(j) = -0.05;
               Ub2(j) = 0.05;
            end
            if Lb2(j) < Lb(j), Lb2(j) = Lb(j); end
            if Ub2(j) > Ub(j), Ub2(j) = Ub(j); end
          end
          s = ' ';
          for i2=1:km
            for i1=1:nVar
              rpop = gBestX;
              rpop(i1) = Lb2(i1) + (Ub2(i1) - Lb2(i1)) * rand;
              zfx = fx(rpop);
              if zfx < gBestCost
                gBestCost = zfx;
                gBestX = rpop;
                s = '*';
                randSearchSuccessCount = randSearchSuccessCount + 1;
              end
            end
          end
          fprintf('Iter = %i, ', iter);
          if ~strcmp(s,'*')
            fprintf('Best Fx = %e, ', gBestCost); 
          else
           fprintf('Best Fx = %e %s, ', gBestCost, s);
          end
          fprintf('Best X = [');
          fprintf('%f, ', gBestX(1:nVar-1));
          fprintf(' %f]\n', gBestX(nVar));

        end
      end
    end
    w = w * wdamp;
  end
  fprintf("random Search Success Count = %d\n", randSearchSuccessCount)
end

